package com.fjhx.email.service.impl;

import cn.hutool.core.date.DateUtil;
import cn.hutool.core.io.IORuntimeException;
import cn.hutool.core.thread.ThreadUtil;
import cn.hutool.core.util.ObjectUtil;
import com.alibaba.fastjson2.JSON;
import com.fjhx.email.config.exception.ServiceException;
import com.fjhx.email.entity.EnterpriseFolder;
import com.fjhx.email.entity.EnterpriseMessage;
import com.fjhx.email.entity.PersonalFolder;
import com.fjhx.email.entity.PersonalMessage;
import com.fjhx.email.entity.dto.MailFolderInfo;
import com.fjhx.email.entity.dto.MailInfo;
import com.fjhx.email.entity.dto.MailSyncInfo;
import com.fjhx.email.entity.dto.MailboxInfo;
import com.fjhx.email.service.*;
import com.fjhx.email.utils.HxHttpUtil;
import com.fjhx.email.utils.email.ImapUtil;
import com.sun.mail.imap.IMAPFolder;
import com.sun.mail.imap.IMAPMessage;
import com.sun.mail.imap.IMAPStore;
import com.sun.mail.util.MailConnectException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.ApplicationArguments;
import org.springframework.boot.ApplicationRunner;
import org.springframework.stereotype.Service;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.TransactionStatus;

import javax.mail.*;
import javax.mail.internet.InternetAddress;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

@Slf4j
@Service
public class CoreServiceImpl implements ApplicationRunner {

    /**
     * 正在同步的邮箱集合
     */
    private static final List<Long> syncMailList = Collections.synchronizedList(new ArrayList<>());

    /**
     * 累计同步次数
     */
    private static final AtomicInteger numCount = new AtomicInteger();

    /**
     * 单次同步邮件总次数
     */
    private static final AtomicInteger mailCount = new AtomicInteger();

    /**
     * 抓取邮件线程池
     */
    private static final Executor mailExecutor = new ThreadPoolExecutor(
            12,
            36,
            60,
            TimeUnit.SECONDS,
            new ArrayBlockingQueue<>(36),
            (runnable, executor) -> {
                if (!executor.isShutdown()) {
                    // 等待0.5秒之后重新入线程池
                    ThreadUtil.sleep(500);
                    executor.execute(runnable);
                }
            }
    );

    @Autowired
    private PlatformTransactionManager platformTransactionManager;

    @Autowired
    private TransactionDefinition transactionDefinition;

    @Autowired
    private IMailService mailService;

    @Autowired
    private IPersonalFolderService personalFolderService;

    @Autowired
    private IPersonalMessageService personalMessageService;

    @Autowired
    private IEnterpriseFolderService enterpriseFolderService;

    @Autowired
    private IEnterpriseMessageService enterpriseMessageService;

    @Override
    public void run(ApplicationArguments args) {

        try {
            List<Long> onlineUserIdList = HxHttpUtil.getOnlineUserIdList();
            MailSyncInfo.mailboxInfoList = mailService.getMailboxInfoListByUserId(onlineUserIdList, MailSyncInfo.mailType);
        } catch (IORuntimeException e) {
            log.error("获取主服务在线用户超时");
            ThreadUtil.sleep(1000 * 5);
            run(args);
            return;
        } catch (ServiceException e) {
            log.error("主服务返回信息错误:{}", e.getMessage());
            ThreadUtil.sleep(1000 * 30);
            run(args);
            return;
        } catch (Exception e) {
            log.error("其他错误", e);
            ThreadUtil.sleep(1000 * 30);
            run(args);
            return;
        }

        // 发起邮件同步
        new Thread(this::monitorMail).start();
    }

    /**
     * 邮件同步
     */
    public void monitorMail() {
        try {
            doMonitorMail();
        } catch (Exception e) {
            log.error("未知异常", e);
        }
        new Thread(this::monitorMail).start();
    }

    /**
     * 执行邮件同步逻辑
     */
    private void doMonitorMail() {

        // 所有需要同步的邮件
        List<MailboxInfo> mailboxInfoList = new ArrayList<>(MailSyncInfo.mailboxInfoList);

        // 所有需要同步的邮件数量
        int mailboxSize = mailboxInfoList.size();

        if (mailboxSize == 0) {
            ThreadUtil.sleep(MailSyncInfo.minWaitingTime);
            return;
        }

        // 开始处理时间
        long start = System.currentTimeMillis();

        // 清空抓取邮件数
        mailCount.set(0);

        // 累计同步次数
        log.info("第 {} 伦邮件同步开始，需要同步 {} 个邮箱", numCount.incrementAndGet(), mailboxSize);

        // 定义一个线程计数器
        CountDownLatch countDownLatch = new CountDownLatch(mailboxSize);

        // 开启抓取邮件任务
        for (MailboxInfo mailbox : mailboxInfoList) {
            // 开启异步操作
            mailExecutor.execute(() -> {
                synchronousMail(mailbox);
                countDownLatch.countDown();
            });
        }

        try {
            boolean awaitFlag = countDownLatch.await(MailSyncInfo.awaitTimeout, TimeUnit.MINUTES);
            if (!awaitFlag) {
                log.error("邮件同步超时执行");
            }
        } catch (InterruptedException e) {
            log.error("countDownLatch.await() 发生异常", e);
        }

        // 结束处理时间
        long end = System.currentTimeMillis();

        // 处理时间
        long handleTime = end - start;

        log.info("第 {} 伦邮件同步完成，已同步 {} 封邮件，总耗时 {}",
                numCount.get(), mailCount.get(), DateUtil.formatBetween(handleTime));

        // 处理时间是否小于最小处理时间
        if (MailSyncInfo.minWaitingTime > handleTime) {
            ThreadUtil.sleep(MailSyncInfo.minWaitingTime - handleTime);
        }

    }

    /**
     * 同步邮箱
     */
    private void synchronousMail(MailboxInfo mailboxInfo) {

        Long id = mailboxInfo.getId();

        if (syncMailList.contains(id)) {
            log.error("邮箱 {} 同步超时：{}", mailboxInfo.getMailUser(), JSON.toJSONString(mailboxInfo));
            return;
        }

        // 开始同步，添加id
        syncMailList.add(id);

        IMAPStore store = null;

        try {

            // 获取邮件连接
            store = ImapUtil.getStore(
                    mailboxInfo.getReceiveHost(),
                    mailboxInfo.getMailUser(),
                    mailboxInfo.getMailPassword());

            // 邮件文件夹
            List<MailFolderInfo> mailFolderInfoList = mailboxInfo.getMailFolderInfoList();

            // 分别拉取每个文件夹邮件
            for (MailFolderInfo mailFolder : mailFolderInfoList) {

                // 获取文件夹
                IMAPFolder folder = (IMAPFolder) store.getFolder(mailFolder.getName());

                // 只读方式打开收件箱
                try {
                    folder.open(Folder.READ_ONLY);
                } catch (FolderNotFoundException folderNotFoundException) {
                    continue;
                }

                try {
                    // 获取文件夹邮件
                    Message[] messages = folder.getMessages();
                    if (messages == null || messages.length == 0) {
                        continue;
                    }

                    // 拉取邮件
                    List<MailInfo> mailInfoList = getMailInfoList(folder, mailFolder, messages);

                    if (ObjectUtil.isEmpty(mailInfoList)) {
                        continue;
                    }

                    // 保存邮件数据，更新文件夹最后同步邮件时间
                    saveMailInfo(mailboxInfo, mailFolder, mailInfoList);

                    // 最后一次同步邮件信息
                    MailInfo mailInfo = mailInfoList.get(mailInfoList.size() - 1);
                    mailFolder.setLastUid(mailInfo.getUid());
                    mailFolder.setLastReceivedDate(mailInfo.getReceivedDate());

                } finally {
                    ImapUtil.closeFolder(folder);
                }
            }

        } catch (AuthenticationFailedException e) {
            log.error("邮箱认证失败：{}", JSON.toJSONString(mailboxInfo), e);
        } catch (MailConnectException e) {
            log.error("邮件连接失败：{}", JSON.toJSONString(mailboxInfo), e);
        } catch (Exception e) {
            log.error("邮件同步出错，邮箱信息：{}", JSON.toJSONString(mailboxInfo), e);
        } finally {
            // 同步完成，删除id
            syncMailList.remove(id);
            try {
                ImapUtil.closeStore(store);
            } catch (MessagingException e) {
                log.error("email链接关闭失败：{}", mailboxInfo);
            }
        }

    }

    /**
     * 拉取邮件
     */
    private List<MailInfo> getMailInfoList(IMAPFolder folder, MailFolderInfo mailFolder, Message[] messages) throws MessagingException {

        // 上次同步邮件uid
        Long lastUid = mailFolder.getLastUid();

        // 上次同步邮件接收时间
        Date lastReceivedDate = mailFolder.getLastReceivedDate();

        // 初始化时最多只同步 initDay 天数据
        if (lastReceivedDate == null) {
            lastReceivedDate = DateUtil.offsetDay(new Date(), -MailSyncInfo.initDay);
        }

        // 收件列表
        List<MailInfo> mailInfoList = new ArrayList<>();

        for (int i = messages.length - 1; i >= 0; i--) {

            // 邮件
            IMAPMessage message = (IMAPMessage) messages[i];

            // 邮件接收时间
            Date receivedDate;
            try {
                receivedDate = message.getReceivedDate();
            } catch (MessagingException messagingException) {
                continue;
            }

            // 邮件uid
            long uid = folder.getUID(message);

            // 最后一封邮件uid相同，代表上次同步邮件到这，不继续找之前的邮件了
            if (lastUid != null && Objects.equals(uid, lastUid)) {
                break;
            }

            // 收件时间在上次收件时间之前，不继续找之前的邮件了
            if (receivedDate.before(lastReceivedDate)) {
                break;
            }

            // 邮件信息
            MailInfo mailInfo = new MailInfo();
            mailInfo.setUid(folder.getUID(message));
            mailInfo.setSubject(message.getSubject());
            mailInfo.setSendDate(message.getSentDate());
            mailInfo.setReceivedDate(receivedDate);

            // 保存发件人信息
            InternetAddress sender = (InternetAddress) message.getSender();
            if (sender != null) {
                mailInfo.setFromEmail(sender.getAddress());
                mailInfo.setFromPersonalName(sender.getPersonal());
            }else {
                Address[] from = message.getFrom();
                if (ObjectUtil.isNotEmpty(from)) {
                    InternetAddress formItem = (InternetAddress) from[0];
                    mailInfo.setFromEmail(formItem.getAddress());
                    mailInfo.setFromPersonalName(formItem.getPersonal());
                }
            }

            // 添加到邮件列表
            mailInfoList.add(0, mailInfo);

            // 添加同步邮件总次数
            mailCount.incrementAndGet();

        }

        return mailInfoList;
    }

    /**
     * 保存邮件数据，更新文件夹最后同步邮件时间
     */
    private void saveMailInfo(MailboxInfo mailboxInfo, MailFolderInfo mailFolder, List<MailInfo> mailInfoList) {

        // 获取抓取的最后一封邮件
        MailInfo lastMailInfo = mailInfoList.get(mailInfoList.size() - 1);

        // 开启事务
        TransactionStatus transactionStatus = platformTransactionManager.getTransaction(transactionDefinition);

        try {

            // 个人邮箱
            if (mailboxInfo.getType() == 1) {

                PersonalFolder personalFolder = new PersonalFolder();
                personalFolder.setId(mailFolder.getId());
                personalFolder.setLastUid(lastMailInfo.getUid());
                personalFolder.setLastReceivedDate(lastMailInfo.getReceivedDate());
                personalFolderService.updateById(personalFolder);

                List<PersonalMessage> personalMessageList = mailInfoList.stream().map(mailInfo -> {
                    PersonalMessage personalMessage = new PersonalMessage();
                    personalMessage.setUid(mailInfo.getUid());
                    personalMessage.setMailboxId(mailboxInfo.getId());
                    personalMessage.setFolderId(mailFolder.getId());
                    personalMessage.setFolderName(mailFolder.getName());
                    personalMessage.setSubject(mailInfo.getSubject());
                    personalMessage.setFromEmail(mailInfo.getFromEmail());
                    personalMessage.setFromPersonalName(mailInfo.getFromPersonalName());
                    personalMessage.setSendDate(mailInfo.getSendDate());
                    personalMessage.setReceivedDate(mailInfo.getReceivedDate());
                    personalMessage.setSyncStatus(0);
                    personalMessage.setDelFlag(0);
                    return personalMessage;
                }).collect(Collectors.toList());
                personalMessageService.saveBatch(personalMessageList);

            }
            // 企业邮箱
            else {

                EnterpriseFolder enterpriseFolder = new EnterpriseFolder();
                enterpriseFolder.setId(mailFolder.getId());
                enterpriseFolder.setLastUid(lastMailInfo.getUid());
                enterpriseFolder.setLastReceivedDate(lastMailInfo.getReceivedDate());
                enterpriseFolderService.updateById(enterpriseFolder);

                List<EnterpriseMessage> enterpriseMessageList = mailInfoList.stream().map(mailInfo -> {
                    EnterpriseMessage enterpriseMessage = new EnterpriseMessage();
                    enterpriseMessage.setUid(mailInfo.getUid());
                    enterpriseMessage.setMailboxId(mailboxInfo.getId());
                    enterpriseMessage.setFolderId(mailFolder.getId());
                    enterpriseMessage.setFolderName(mailFolder.getName());
                    enterpriseMessage.setSubject(mailInfo.getSubject());
                    enterpriseMessage.setFromEmail(mailInfo.getFromEmail());
                    enterpriseMessage.setFromPersonalName(mailInfo.getFromPersonalName());
                    enterpriseMessage.setSendDate(mailInfo.getSendDate());
                    enterpriseMessage.setReceivedDate(mailInfo.getReceivedDate());
                    enterpriseMessage.setSyncStatus(0);
                    enterpriseMessage.setDelFlag(0);
                    return enterpriseMessage;
                }).collect(Collectors.toList());
                enterpriseMessageService.saveBatch(enterpriseMessageList);

            }

            // 提交事务
            platformTransactionManager.commit(transactionStatus);

        } catch (Exception e) {

            // 回滚事务
            platformTransactionManager.rollback(transactionStatus);

            throw new RuntimeException("保存邮件失败", e);

        }

    }

}
